from django.utils import timezone
from djes.models import Indexable
from elasticsearch_dsl import aggs
from elasticsearch_dsl.filter import Term, Range, MatchAll, Nested, Missing


class Searchable(Indexable):
    """adds a `.search` class method to the model
    """

    class Meta(object):
        abstract = True

    @staticmethod
    def _handle_status_filter(status):
        """builds a filter around the `published` field based on a given status

        :param status: published|scheduled|draft indicator
        :type status: str

        :return: a filter to set a time range around `published`
        :rtype: elasticsearch_dsl.filter.F
        """
        # build empty filter; set epoch
        f = MatchAll()
        now = timezone.now()

        # match published
        if status.lower() == "published":
            f &= Range(published={"lte": now})

        # match scheduled
        elif status.lower() == "scheduled":
            f &= Range(published={"gte": now})

        # match draft
        elif status.lower() == "draft":
            f &= Missing(field="published")

        # done
        return f

    @classmethod
    def _build_filters(cls, filters):
        """builds filters for all passed key-values

        :param filters: key-value pairs of field name keys and filter term values
        :type filters: dict

        :return: a compounded filter
        :rtype: elasticsearch_dsl.filter.F
        """
        # build empty filter; iterate filters dict
        f = MatchAll()
        for key, value in filters.items():

            # handle status meta filtering
            if key.lower() == "status":
                status_filter = cls._handle_status_filter(value)
                f &= status_filter

            # handle nested filtering -- convert dunders to dots
            elif "__" in key or "." in key:
                nested_key = key.lower().replace("__", ".")
                path = nested_key.split(".")[0]
                f &= Nested(path=path, filter=Term(**{nested_key: value}))

            # regular term filter
            else:
                key = key.lower()
                f &= Term(**{key: value})

        # done
        return f

    @classmethod
    def _build_ordering(cls, ordering):
        """builds the ordering list and properly converts dunders to dots for nested fields

        :param ordering: field names used to order the results
        :type ordering: list

        :return: a proper list of field names
        :rtype: list
        """
        if isinstance(ordering, str):
            ordering = [ordering, ]

        formatted = []
        for key in ordering:
            nested_key = key.lower().replace("__", ".")
            formatted.append(nested_key)

        return formatted

    @classmethod
    def search(cls, query=None, filters=None, ordering=None):
        """performs a query using the model's `.search_objects` manager

        :param query: terms used to perform query
        :type query: str

        :param filters: key-value pairs used to build filters to limit search results
        :type filters: dict

        :param ordering: field names used to order the results
        :type ordering: list

        :return: elasticsearch search results mapped to django model proxies
        :rtype: django.db.models.QuerySet
        """
        # build initial query set
        qs = cls.search_objects.search()

        # add query if exists
        if query:
            qs = qs.query("match", _all=query)

        # add filters if exist
        if filters:
            built_filters = cls._build_filters(filters)
            qs = qs.filter(built_filters)

        # add ordering if exists
        if ordering:
            ordering = cls._build_ordering(ordering)
            qs = qs.sort(*ordering)

        # done
        return qs


class Aggregateable(Searchable):
    """extends the Searchable model type by adding a `.get_aggregates` class method to the model
    """

    class Meta(object):
        abstract = True

    class Aggregates(object):
        """this is where you need to load up your aggregates
        """
        # channel = {
        #     "path": "channel",
        #     "field": "channel.name.raw"
        # }
        pass

    @classmethod
    def _get_aggregate_declarations(cls):
        """parsed an internal Aggregates subclass to help build aggregate declarations

        :return: a list of bucket names and mapping dictionaries tuples
        :rtype: list
        """
        # check for an Aggregates subclass
        if not hasattr(cls, "Aggregates"):
            raise Exception("You must explicitly set an `Aggregates` subclass")

        # get fields
        agg_mapping_fields = [field for field in dir(cls.Aggregates) if not field.startswith("_")]

        # init container; iterate field names; append to container
        agg_mappings = []
        for field in agg_mapping_fields:
            declarations = getattr(cls.Aggregates, field)
            agg_mappings.append((field, declarations))

        # done
        return agg_mappings

    @classmethod
    def _build_aggregates(cls, qs):
        """builds aggregates onto a queryset

        :param qs: elasticsearch search results mapped to django model proxies
        :type qs: django.db.models.QuerySet

        :return: the updated query, a list of bucket names, and a list of paths
        :rtype: tuple
        """
        # get declarations; init containers
        agg_mappings = cls._get_aggregate_declarations()
        names = []
        fields = []

        # iterate declarations
        for name, mapping in agg_mappings:
            try:
                # parse mapping
                path = mapping["path"]
                field = mapping["field"]

                # bolt on aggregate to query
                qs.aggs.bucket(
                    name,
                    aggs.Nested(
                        path=path,
                        aggs={path: aggs.Terms(field=field)}
                    )
                )

                # append values to parse later
                names.append(name)
                fields.append(field)

            # they messed up the subclass
            except KeyError:
                raise Exception("Misconfigured aggregate declaration: {}, {}".format(name, mapping))

        # done
        return qs, names, fields

    @classmethod
    def get_aggregates(cls, query=None, filters=None):
        """performs an aggregation query using the model's `.search` class method

        :param query: terms used to perform query
        :type query: str

        :param filters: key-value pairs used to build filters to limit search results
        :type filters: dict

        :return: a dictionary of field keys and value/count mapped dictionary values
        :rtype: dict
        """
        # get initial query set
        qs = cls.search(query, filters)

        # build aggregates
        qs, buckets, fields = cls._build_aggregates(qs)

        # execute
        raw_aggregates = qs.execute().aggregations

        # parse
        aggregates = {}
        for index, field in enumerate(fields):
            dunder_field = field.lower().replace(".", "__")
            if filters:
                if dunder_field in filters:
                    continue
            bucket = buckets[index]
            parsed_aggregates = dict([(b["key"], b["doc_count"]) for b in raw_aggregates[bucket][bucket]["buckets"]])
            aggregates[dunder_field] = parsed_aggregates

        # done
        return aggregates



from copy import deepcopy

from rest_framework import viewsets, status
from rest_framework.decorators import list_route
from rest_framework.response import Response

from djesrf.models import Searchable, Aggregateable


class SearchableModelViewSet(viewsets.ModelViewSet):
    """moves all list functionality to Elasticsearch and off the ORM
    """

    model = object

    def __init__(self, **kwargs):
        if not issubclass(self.model, Searchable):
            raise Exception("You must explicitly supply a `model` attribute of this viewset "
                            "and it must subclass `djesrf.models.Searchable`")
        super(SearchableModelViewSet, self).__init__(**kwargs)

    def list(self, request, *args, **kwargs):
        # get params
        params = deepcopy(request.query_params)

        if "search" in params:
            query = params["search"]
            del params["search"]
        else:
            query = None

        if "ordering" in params:
            ordering = params.getlist("ordering")
            del params["ordering"]
        else:
            ordering = None

        if "page" in params:
            del params["page"]

        if "page_size" in params:
            del params["page_size"]

        results = self.model.search(query, params, ordering)

        page = self.paginate_queryset(results)
        if page:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)
        serializer = self.get_serializer(results, many=True)
        return Response(serializer.data)


class AggregateableModelViewSet(SearchableModelViewSet):

    def __init__(self, **kwargs):
        if not issubclass(self.model, Aggregateable):
            raise Exception("You must explicitly supply a `model` attribute of this viewset "
                            "and it must subclass `djesrf.models.Aggregateable`")
        super(AggregateableModelViewSet, self).__init__(**kwargs)

    @list_route(methods=["get"])
    def aggregates(self, request):
        # get params
        params = deepcopy(request.QUERY_PARAMS)

        if "search" in params:
            query = params["search"]
            del params["search"]
        else:
            query = None

        if "ordering" in params:
            del params["ordering"]

        if "page" in params:
            del params["page"]

        if "page_size" in params:
            del params["page_size"]

        results = self.model.get_aggregates(query, params)

        response = {
            "count": len(results),
            "next": None,
            "previous": None,
            "results": [],
        }

        for path, obj in results.items():
            name = path.split(".")[0].title()
            path = path.replace(".", "__")
            result = {
                "name": name,
                "path": path,
                "aggregates": [],
            }
            for value, count in obj.items():
                result["aggregates"].append({"value": value, "count": count})
            response["results"].append(result)
        return Response(response)
